import { useReducer } from 'react';

interface State<T> {
  stack: T[];
}
interface Action<T> {
  type: 'push' | 'pop' | 'clear' | 'reset';
  values?: T[];
}

function createReducer<T>() {
  return function (state: State<T>, action: Action<T>) {
    switch (action.type) {
      case 'push': {
        state.stack.push(...action.values!);
        break;
      }
      case 'pop': {
        state.stack.pop();
        break;
      }
      case 'clear': {
        return { stack: [] };
      }
      case 'reset': {
        return { stack: action.values || [] };
      }
      default:
        break;
    }
    return {
      stack: state.stack,
    };
  };
}

export function useStack<T>(arr: T[] = []) {
  const [state, dispatch] = useReducer(createReducer<T>(), {
    stack: arr,
  });

  return {
    push: (...values: T[]) => {
      dispatch({ type: 'push', values });
      return values;
    },
    pop: () => {
      const item = state.stack[state.stack.length - 1] || null;
      dispatch({ type: 'pop' });
      return item;
    },
    clear: () => dispatch({ type: 'clear' }),
    reset: (values: T[] = []) => {
      dispatch({ type: 'reset', values });
    },
    forEach: (callback: (value: T, index: number, arr: T[]) => void) =>
      state.stack.reverse().forEach(callback),
    map: function <Result>(
      callback: (value: T, index: number, arr: T[]) => Result,
    ) {
      return state.stack.reverse().map(callback);
    },
    includes: (value: T) => state.stack.includes(value),
    peek: () => state.stack[state.stack.length - 1] || null,
    entries: () => state.stack.entries(),
    values: () => state.stack.values(),
    keys: () => state.stack.keys(),
    size: state.stack.length,
    *[Symbol.iterator]() {
      const iterator1 = state.stack[Symbol.iterator]();
      for (const item of iterator1) {
        yield item;
      }
    },
  };
}
